import warnings
from collections.abc import Callable, Collection, Hashable, Iterable
from dataclasses import dataclass, field
from pathlib import Path
from typing import Generic, Literal

from cognite.client import data_modeling as dm
from cognite.client.data_classes._base import (
    T_CogniteResourceList,
    T_WritableCogniteResource,
    T_WriteClass,
)
from cognite.client.data_classes.data_modeling import (
    DataModelApplyList,
    DataModelId,
    SpaceApply,
    ViewApplyList,
)
from cognite.client.exceptions import CogniteAPIError

from cognite.neat._client import DataModelingLoader, NeatClient
from cognite.neat._client._api.data_modeling_loaders import MultiCogniteAPIError, T_WritableCogniteResourceList
from cognite.neat._client.data_classes.data_modeling import Component, ViewApplyDict
from cognite.neat._client.data_classes.schema import DMSSchema
from cognite.neat._issues import IssueList
from cognite.neat._issues.warnings import (
    PrincipleOneModelOneSpaceWarning,
    ResourceRetrievalWarning,
)
from cognite.neat._rules.models.dms import DMSRules
from cognite.neat._shared import T_ID
from cognite.neat._utils.upload import UploadResult

from ._base import CDFExporter


@dataclass
class ItemCategorized(Generic[T_ID, T_WriteClass]):
    resource_name: str
    as_id: Callable[[T_WriteClass], T_ID]
    to_create: list[T_WriteClass] = field(default_factory=list)
    to_update: list[T_WriteClass] = field(default_factory=list)
    to_delete: list[T_WriteClass] = field(default_factory=list)
    to_skip: list[T_WriteClass] = field(default_factory=list)
    unchanged: list[T_WriteClass] = field(default_factory=list)

    @property
    def to_create_ids(self) -> list[T_ID]:
        return [self.as_id(item) for item in self.to_create]

    @property
    def to_update_ids(self) -> list[T_ID]:
        return [self.as_id(item) for item in self.to_update]

    @property
    def to_skip_ids(self) -> list[T_ID]:
        return [self.as_id(item) for item in self.to_skip]

    @property
    def to_delete_ids(self) -> list[T_ID]:
        return [self.as_id(item) for item in self.to_delete]

    @property
    def unchanged_ids(self) -> list[T_ID]:
        return [self.as_id(item) for item in self.unchanged]

    def item_ids(self) -> Iterable[T_ID]:
        yield from (self.as_id(item) for item in self.to_create + self.to_update + self.to_delete + self.unchanged)


class DMSExporter(CDFExporter[DMSRules, DMSSchema]):
    """Export rules to Cognite Data Fusion's Data Model Storage (DMS) service.

    Args:
        export_components (frozenset[Literal["all", "spaces", "data_models", "views", "containers"]], optional):
            Which components to export. Defaults to frozenset({"all"}).
        include_space (set[str], optional):
            If set, only export components in the given spaces. Defaults to None which means all spaces.
        existing (Literal["fail", "skip", "update", "force"], optional): How to handle existing components.
            Defaults to "update". See below for details.
        instance_space (str, optional): The space to use for the instance. Defaults to None.
        suppress_warnings (bool, optional): Suppress warnings. Defaults to False.
        remove_cdf_spaces (bool, optional): Skip views and containers that are system are in system spaces.

    ... note::

        - "fail": If any component already exists, the export will fail.
        - "skip": If any component already exists, it will be skipped.
        - "update": If any component already exists, it will
        - "force": If any component already exists, it will be deleted and recreated.

    """

    def __init__(
        self,
        export_components: Component | Collection[Component] | None = None,
        include_space: set[str] | None = None,
        existing: Literal["fail", "skip", "update", "force", "recreate"] = "update",
        instance_space: str | None = None,
        suppress_warnings: bool = False,
        drop_data: bool = False,
        remove_cdf_spaces: bool = True,
    ):
        self.export_components = export_components
        self.include_space = include_space
        self.existing = existing
        self.drop_data = drop_data
        self.instance_space = instance_space
        self.suppress_warnings = suppress_warnings
        self._schema: DMSSchema | None = None
        self.remove_cdf_spaces = remove_cdf_spaces

    @property
    def description(self) -> str:
        return "Export verified DMS Model to CDF."

    def export_to_file(self, rules: DMSRules, filepath: Path) -> None:
        """Export the rules to a file(s).

        If the file is a directory, the components will be exported to separate files, otherwise they will be
        exported to a zip file.

        Args:
            filepath: Directory or zip file path to export to.
            rules:
        """
        if filepath.is_dir():
            self._export_to_directory(filepath, rules)
        else:
            self._export_to_zip_file(filepath, rules)

    def _export_to_directory(self, directory: Path, rules: DMSRules) -> None:
        schema = self.export(rules)
        exclude = self._create_exclude_set()
        schema.to_directory(directory, exclude=exclude, new_line=self._new_line, encoding=self._encoding)

    def _export_to_zip_file(self, filepath: Path, rules: DMSRules) -> None:
        if filepath.suffix not in {".zip"}:
            warnings.warn("File extension is not .zip, adding it to the file name", stacklevel=2)
            filepath = filepath.with_suffix(".zip")
        schema = self.export(rules)
        exclude = self._create_exclude_set()
        schema.to_zip(filepath, exclude=exclude)

    def _create_exclude_set(self):
        if self.export_components is None:
            exclude = set()
        else:
            exclude = {"spaces", "data_models", "views", "containers", "node_types"} - set(self.export_components)
        return exclude

    def export(self, rules: DMSRules) -> DMSSchema:
        # We do not want to include CogniteCore/CogniteProcess Industries in the schema
        return rules.as_schema(instance_space=self.instance_space, remove_cdf_spaces=self.remove_cdf_spaces)

    def delete_from_cdf(
        self, rules: DMSRules, client: NeatClient, dry_run: bool = False, skip_space: bool = False
    ) -> Iterable[UploadResult]:
        schema = self.export(rules)

        # we need to reverse order in which we are picking up the items to delete
        # as they are sorted in the order of creation and we need to delete them in reverse order
        for loader in reversed(client.loaders.by_dependency_order(self.export_components)):
            items = loader.items_from_schema(schema)
            item_ids = loader.get_ids(items)
            existing_items = loader.retrieve(item_ids)
            existing_ids = set(loader.get_ids(existing_items))
            to_delete: list[Hashable] = []
            for item_id in item_ids:
                if (
                    isinstance(loader, DataModelingLoader)
                    and self.include_space is not None
                    and not loader.in_space(item_id, self.include_space)
                ):
                    continue

                if item_id in existing_ids:
                    to_delete.append(item_id)

            result = UploadResult(loader.resource_name)  # type: ignore[var-annotated]
            if dry_run:
                result.deleted.update(to_delete)
                yield result
                continue

            if to_delete:
                try:
                    deleted = loader.delete(to_delete)
                except MultiCogniteAPIError as e:
                    result.deleted.update([loader.get_id(item) for item in e.success])
                    result.failed_deleted.update([loader.get_id(item) for item in e.failed])
                    for error in e.errors:
                        result.error_messages.append(f"Failed to delete {loader.resource_name}: {error!s}")
                else:
                    result.deleted.update(deleted)
            yield result

    def export_to_cdf_iterable(
        self, rules: DMSRules, client: NeatClient, dry_run: bool = False
    ) -> Iterable[UploadResult]:
        schema = self.export(rules)

        # The CDF UI does not deal well with a child view overwriting a parent property with the same name
        # This is a workaround to remove the duplicated properties
        self._remove_duplicated_properties(schema.views, client)

        categorized_items_by_loader = self._categorize_by_loader(client, schema)

        is_failing = self.existing == "fail" and any(
            loader.resource_name for loader, categorized in categorized_items_by_loader.items() if categorized.to_update
        )

        deleted_by_name: dict[str, UploadResult] = {}
        if not is_failing:
            # Deletion is done in reverse order to take care of dependencies
            for loader, items in reversed(categorized_items_by_loader.items()):
                issue_list = IssueList()

                if items.resource_name == client.loaders.data_models.resource_name:
                    warning_list = self._validate(list(items.item_ids()), client)
                    issue_list.extend(warning_list)

                results = UploadResult(loader.resource_name, issues=issue_list)  # type: ignore[var-annotated]
                if dry_run:
                    results.deleted.update(items.to_delete_ids)
                else:
                    if items.to_delete_ids:
                        try:
                            deleted = loader.delete(items.to_delete_ids)
                        except MultiCogniteAPIError as e:
                            results.deleted.update([loader.get_id(item) for item in e.success])
                            results.failed_deleted.update([loader.get_id(item) for item in e.failed])
                            for error in e.errors:
                                results.error_messages.append(f"Failed to delete {loader.resource_name}: {error!s}")
                        else:
                            results.deleted.update(deleted)
                deleted_by_name[loader.resource_name] = results

        for loader, items in categorized_items_by_loader.items():
            issue_list = IssueList()

            if items.resource_name == client.loaders.data_models.resource_name:
                warning_list = self._validate(list(items.item_ids()), client)
                issue_list.extend(warning_list)

            results = UploadResult(loader.resource_name, issues=issue_list)  # type: ignore[var-annotated]
            if is_failing:
                # If any component already exists, the export will fail.
                # This is the same if we run dry_run or not.
                results.failed_upserted.update(items.to_update_ids)
                results.failed_created.update(items.to_create_ids)
                results.failed_deleted.update(items.to_delete_ids)
                results.unchanged.update(items.unchanged_ids)
                results.error_messages.append("Existing components found and existing_handling is 'fail'")
                yield results
                continue

            results.unchanged.update(items.unchanged_ids)
            results.skipped.update(items.to_skip_ids)
            if delete_results := deleted_by_name.get(loader.resource_name):
                results.deleted.update(delete_results.deleted)
                results.failed_deleted.update(delete_results.failed_deleted)
                results.error_messages.extend(delete_results.error_messages)

            if dry_run:
                if self.existing in ["update", "force"]:
                    # Assume all changed are successful
                    results.changed.update(items.to_update_ids)
                elif self.existing == "skip":
                    results.skipped.update(items.to_update_ids)
                results.created.update(items.to_create_ids)
                yield results
                continue

            if items.to_create:
                try:
                    created = loader.create(items.to_create)
                except MultiCogniteAPIError as e:
                    results.created.update([loader.get_id(item) for item in e.success])
                    results.failed_created.update([loader.get_id(item) for item in e.failed])
                    for error in e.errors:
                        results.error_messages.append(f"Failed to create {loader.resource_name}: {error!s}")
                else:
                    results.created.update(loader.get_ids(created))

            if items.to_update and self.existing == "skip":
                results.skipped.update(items.to_update_ids)
            elif items.to_update:
                try:
                    updated = loader.update(items.to_update, force=self.existing == "force", drop_data=self.drop_data)
                except MultiCogniteAPIError as e:
                    results.changed.update([loader.get_id(item) for item in e.success])
                    results.failed_changed.update([loader.get_id(item) for item in e.failed])
                    for error in e.errors:
                        results.error_messages.append(f"Failed to update {loader.resource_name}: {error!s}")
                else:
                    results.changed.update(loader.get_ids(updated))

            yield results

    def _categorize_by_loader(self, client: NeatClient, schema: DMSSchema) -> dict[DataModelingLoader, ItemCategorized]:
        categorized_items_by_loader: dict[DataModelingLoader, ItemCategorized] = {}
        redeploy_data_model = False
        for loader in client.loaders.by_dependency_order(self.export_components):
            items = loader.items_from_schema(schema)
            # The conversion from DMS to GraphQL does not seem to be triggered even if the views
            # are changed. This is a workaround to force the conversion.
            is_redeploying = isinstance(items, DataModelApplyList) and redeploy_data_model

            categorized = self._categorize_items_for_upload(loader, items, is_redeploying)
            categorized_items_by_loader[loader] = categorized

            if isinstance(items, ViewApplyList) and (categorized.to_create or categorized.to_update):
                redeploy_data_model = True
        return categorized_items_by_loader

    def _categorize_items_for_upload(
        self,
        loader: DataModelingLoader[
            T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList
        ],
        items: T_CogniteResourceList,
        is_redeploying: bool,
    ) -> ItemCategorized[T_ID, T_WriteClass]:
        item_ids = loader.get_ids(items)
        cdf_items = loader.retrieve(item_ids)
        cdf_item_by_id = {loader.get_id(item): item for item in cdf_items}
        categorized = ItemCategorized[T_ID, T_WriteClass](loader.resource_name, loader.get_id)
        for item in items:
            if (
                isinstance(items, DataModelApplyList)
                and self.include_space is not None
                and not loader.in_space(item, self.include_space)
            ):
                continue
            item_id = loader.get_id(item)
            cdf_item = cdf_item_by_id.get(item_id)
            if cdf_item is None:
                categorized.to_create.append(item)
            elif (is_redeploying or self.existing == "recreate") and not isinstance(item, SpaceApply):
                # Spaces are not deleted, instead they are updated. Deleting a space is an expensive operation
                # and are seldom needed. If you need to delete the space, it should be done in a different operation.
                if not self.drop_data and loader.has_data(item_id):
                    categorized.to_skip.append(cdf_item)
                else:
                    categorized.to_delete.append(cdf_item.as_write())
                    categorized.to_create.append(item)
            elif loader.are_equal(item, cdf_item):
                categorized.unchanged.append(item)
            else:
                categorized.to_update.append(item)
        return categorized

    def _validate(self, items: list[DataModelId], client: NeatClient) -> IssueList:
        issue_list = IssueList()
        if other_models := self._exist_other_data_models(client, items):
            warning = PrincipleOneModelOneSpaceWarning(
                f"There are multiple data models in the same space {items[0].space}. "
                f"Other data models in the space are {other_models}.",
            )
            if not self.suppress_warnings:
                warnings.warn(warning, stacklevel=2)
            issue_list.append(warning)

        return issue_list

    @classmethod
    def _exist_other_data_models(cls, client: NeatClient, model_ids: list[DataModelId]) -> list[DataModelId]:
        if not model_ids:
            return []
        space = model_ids[0].space
        external_id = model_ids[0].external_id
        try:
            data_models = client.data_modeling.data_models.list(space=space, limit=25, all_versions=False)
        except CogniteAPIError as e:
            warnings.warn(ResourceRetrievalWarning(frozenset({space}), "space", str(e)), stacklevel=2)
            return []
        else:
            return [
                data_model.as_id()
                for data_model in data_models
                if (data_model.space, data_model.external_id) != (space, external_id)
            ]

    @staticmethod
    def _remove_duplicated_properties(views: ViewApplyDict, client: NeatClient) -> None:
        parent_view_ids = {parent for view in views.values() for parent in view.implements}
        parent_view_list = client.data_modeling.views.retrieve(
            list(parent_view_ids), include_inherited_properties=False
        )
        parent_view_by_id = {view.as_id(): view.as_write() for view in parent_view_list}
        for view in views.values():
            if view.implements is None:
                continue
            for parent_id in view.implements:
                if not (parent_view := parent_view_by_id.get(parent_id)):
                    continue
                for shared_prop_id in set(view.properties or {}) & set(parent_view.properties or {}):
                    if view.properties is None or parent_view.properties is None:
                        continue
                    prop = view.properties[shared_prop_id]
                    parent_prop = parent_view.properties[shared_prop_id]
                    if (
                        isinstance(prop, dm.MappedPropertyApply)
                        and isinstance(parent_prop, dm.MappedPropertyApply)
                        and (
                            prop.container_property_identifier == parent_prop.container_property_identifier
                            and prop.container == parent_prop.container
                            and prop.source == parent_prop.source
                        )
                    ):
                        view.properties.pop(shared_prop_id)



"""This is module contains all the Neat Exceptions (Errors) and Warnings as well
as some helper classes to handle them like NeatIssueList"""

from ._base import (
    IssueList,
    MultiValueError,
    NeatError,
    NeatIssue,
    NeatWarning,
)
from ._contextmanagers import catch_issues, catch_warnings

__all__ = [
    "IssueList",
    "MultiValueError",
    "NeatError",
    "NeatIssue",
    "NeatWarning",
    "catch_issues",
    "catch_warnings",
]



from collections.abc import Iterable
from pathlib import Path
from typing import cast

from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, AssetFilter, AssetList

from ._base import ClassicCDFBaseExtractor, InstanceIdPrefix


class AssetsExtractor(ClassicCDFBaseExtractor[Asset]):
    """Extract data from Cognite Data Fusions Assets into Neat."""

    _default_rdf_type = "Asset"
    _instance_id_prefix = InstanceIdPrefix.asset

    @classmethod
    def _from_dataset(cls, client: CogniteClient, data_set_external_id: str) -> tuple[int | None, Iterable[Asset]]:
        total = client.assets.aggregate_count(filter=AssetFilter(data_set_ids=[{"externalId": data_set_external_id}]))
        items = client.assets(data_set_external_ids=data_set_external_id)
        return total, items

    @classmethod
    def _from_hierarchy(cls, client: CogniteClient, root_asset_external_id: str) -> tuple[int | None, Iterable[Asset]]:
        total = client.assets.aggregate_count(
            filter=AssetFilter(asset_subtree_ids=[{"externalId": root_asset_external_id}])
        )
        items = cast(
            Iterable[Asset],
            client.assets(asset_subtree_external_ids=root_asset_external_id),
        )
        return total, items

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[Asset]]:
        assets = AssetList.load(Path(file_path).read_text())
        return len(assets), assets



from collections import defaultdict
from collections.abc import Iterable

from cognite.neat._rules.models.entities import ClassEntity
from cognite.neat._rules.models.information import InformationProperty


def duplicated_properties(
    properties: Iterable[InformationProperty],
) -> dict[tuple[ClassEntity, str], list[tuple[int, InformationProperty]]]:
    class_properties_by_id: dict[tuple[ClassEntity, str], list[tuple[int, InformationProperty]]] = defaultdict(list)
    for prop_no, prop in enumerate(properties):
        class_properties_by_id[(prop.class_, prop.property_)].append((prop_no, prop))
    return {k: v for k, v in class_properties_by_id.items() if len(v) > 1}



import dataclasses
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import ClassVar, TypeAlias, cast

from rdflib import Graph
from rdflib.query import ResultRow

from cognite.neat._issues.warnings import NeatValueWarning
from cognite.neat._shared import Triple
from cognite.neat._utils.collection_ import iterate_progress_bar_if_above_config_threshold
from cognite.neat._utils.graph_transformations_report import GraphTransformationResult

To_Add_Triples: TypeAlias = set[Triple]
To_Remove_Triples: TypeAlias = set[Triple]


@dataclasses.dataclass
class RowTransformationOutput:
    remove_triples: To_Remove_Triples = dataclasses.field(default_factory=set)
    add_triples: To_Add_Triples = dataclasses.field(default_factory=set)
    instances_removed_count: int = 0
    instances_added_count: int = 0
    instances_modified_count: int = 0


class BaseTransformer(ABC):
    description: str
    _use_only_once: bool = False
    _need_changes: ClassVar[frozenset[str]] = frozenset()

    @abstractmethod
    def transform(self, graph: Graph) -> None:
        raise NotImplementedError()


class BaseTransformerStandardised(ABC):
    """Standardised base transformer to use in case a transformer is adding or removing triples from a graph. If you
    are doing more specialised operations, please overwrite the .transform() method.
    """

    description: str
    _use_only_once: bool = False
    _need_changes: ClassVar[frozenset[str]] = frozenset()

    @abstractmethod
    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        """The operations to perform on each row resulting from the ._iterate_query() method.
        The operation should return a list of triples to add and to remove.
        """
        raise NotImplementedError()

    @abstractmethod
    def _count_query(self) -> str:
        """
        Overwrite to fetch all affected properties in the graph as a result of the transformation.
        Returns:
            A query string.
        """
        raise NotImplementedError()

    @abstractmethod
    def _iterate_query(self) -> str:
        """
        The query to use for extracting target triples from the graph and performing the transformation.
        Returns:
            A query string.

        !!! note "Complex Queries"
            In majority of cases the query should be a simple SELECT query. However, in case
            when there is a need to have one or more sub iterators, one can overwrite the ._iterator() method
        """
        raise NotImplementedError()

    def _iterator(self, graph: Graph) -> Iterator:
        yield from graph.query(self._iterate_query())

    def _skip_count_query(self) -> str:
        """
        The query to use for extracting target triples from the graph and performing the transformation.
        Returns:
            A query string.
        """
        return ""

    def transform(self, graph: Graph) -> GraphTransformationResult:
        outcome = GraphTransformationResult(self.__class__.__name__)
        outcome.added = outcome.modified = outcome.removed = 0

        iteration_count_res = list(graph.query(self._count_query()))
        iteration_count = int(iteration_count_res[0][0])  # type: ignore [index, arg-type]

        outcome.affected_nodes_count = iteration_count

        if self._skip_count_query():
            skipped_count_res = list(graph.query(self._skip_count_query()))
            skipped_count = int(skipped_count_res[0][0])  # type: ignore [index, arg-type]
            warnings.warn(
                NeatValueWarning(f"Skipping {skipped_count} properties in transformation {self.__class__.__name__}"),
                stacklevel=2,
            )
            outcome.skipped = skipped_count

        if iteration_count == 0:
            return outcome

        result_iterable = self._iterator(graph)
        result_iterable = iterate_progress_bar_if_above_config_threshold(
            result_iterable, iteration_count, self.description
        )

        for row in result_iterable:
            row = cast(ResultRow, row)
            row_output = self.operation(row)

            outcome.added += row_output.instances_added_count
            outcome.removed += row_output.instances_removed_count
            outcome.modified += row_output.instances_modified_count

            for triple in row_output.add_triples:
                graph.add(triple)
            for triple in row_output.remove_triples:
                graph.remove(triple)

        return outcome



from typing import Any, ClassVar

from pydantic import (
    BaseModel,
    model_serializer,
    model_validator,
)

from cognite.neat._rules._constants import EntityTypes
from cognite.neat._rules.models.data_types import DataType

from ._constants import _PARSE, Undefined
from ._single_value import ClassEntity, UnknownEntity


class MultiValueTypeInfo(BaseModel):
    type_: ClassVar[EntityTypes] = EntityTypes.multi_value_type
    types: list[DataType | ClassEntity]

    def __str__(self) -> str:
        return ", ".join([str(t) for t in self.types])

    @model_serializer(when_used="unless-none", return_type=str)
    def as_str(self) -> str:
        return str(self)

    @classmethod
    def load(cls, data: Any) -> "MultiValueTypeInfo":
        # already instance of MultiValueTypeInfo
        if isinstance(data, cls):
            return data

        # it is a raw string that needs to be parsed
        elif isinstance(data, str):
            return cls.model_validate({_PARSE: data})

        # it is dict that needs to be parsed
        else:
            return cls.model_validate(data)

    @model_validator(mode="before")
    def _load(cls, data: Any) -> "dict | MultiValueTypeInfo":
        if isinstance(data, dict) and _PARSE in data:
            data = data[_PARSE]
        elif isinstance(data, dict):
            return data
        else:
            raise ValueError(f"Cannot load {cls.__name__} from {data}")

        result = cls._parse(data)
        return result

    @classmethod
    def _parse(cls, raw: str) -> dict:
        if not (types := [type_.strip() for type_ in raw.split(",")]):
            return {"types": [UnknownEntity()]}
        else:
            return {
                "types": [
                    (DataType.load(type_) if DataType.is_data_type(type_) else ClassEntity.load(type_))
                    for type_ in types
                ]
            }

    def set_default_prefix(self, prefix: str):
        for type_ in self.types:
            if isinstance(type_, ClassEntity) and type_.prefix is Undefined:
                type_.prefix = prefix

    def is_multi_object_type(self) -> bool:
        """Will signalize to DMS converter to create connection to unknown Node type"""
        return all(isinstance(t, ClassEntity) for t in self.types)

    def is_multi_data_type(self) -> bool:
        """Will signalize to DMS converter to attempt to find the best data type for value"""
        return all(isinstance(t, DataType) for t in self.types)

    def is_mixed_type(self) -> bool:
        """Will signalize to DMS converter to fall back to string"""
        return not self.is_multi_object_type() and not self.is_multi_data_type()

    def __hash__(self) -> int:
        return hash(str(self))



import json
import urllib.parse
from collections.abc import Callable, Iterable, Mapping, Set
from typing import Any

from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import Instance
from rdflib import XSD, Literal, Namespace, URIRef

from cognite.neat._shared import Triple
from cognite.neat._utils.auxiliary import string_to_ideal_type

from ._base import BaseExtractor

DEFAULT_EMPTY_VALUES = frozenset({"nan", "null", "none", "", " ", "nil", "n/a", "na", "unknown", "undefined"})


class DictExtractor(BaseExtractor):
    def __init__(
        self,
        id_: URIRef,
        data: Mapping[str, Any],
        namespace: Namespace,
        uri_ref_keys: set[str] | None = None,
        empty_values: Set[str] = DEFAULT_EMPTY_VALUES,
        str_to_ideal_type: bool = False,
        unpack_json: bool = False,
    ) -> None:
        self.id_ = id_
        self.namespace = namespace
        self.data = data
        self.uri_ref_keys = uri_ref_keys or set()
        self.empty_values = empty_values
        self.str_to_ideal_type = str_to_ideal_type
        self.unpack_json = unpack_json

    def extract(self) -> Iterable[Triple]:
        for key, value in self.data.items():
            for predicate_str, object_ in self._get_predicate_objects_pair(key, value, self.unpack_json):
                yield self.id_, self.namespace[urllib.parse.quote(predicate_str)], object_

    def _get_predicate_objects_pair(
        self, key: str, value: Any, unpack_json: bool
    ) -> Iterable[tuple[str, Literal | URIRef]]:
        if key in self.uri_ref_keys and not isinstance(value, dict | list):
            yield key, URIRef(self.namespace[urllib.parse.quote(value)])
        if isinstance(value, str | float | bool | int):
            yield key, Literal(value)
        elif isinstance(value, dict) and unpack_json:
            yield from self._unpack_json(value)
        elif isinstance(value, dict):
            # This object is a json object.
            yield key, Literal(json.dumps(value), datatype=XSD._NS["json"])
        elif isinstance(value, list):
            for item in value:
                yield from self._get_predicate_objects_pair(key, item, False)

    def _unpack_json(self, value: dict, parent: str | None = None) -> Iterable[tuple[str, Literal | URIRef]]:
        for sub_key, sub_value in value.items():
            key = f"{parent}_{sub_key}" if parent else sub_key
            if isinstance(sub_value, str):
                if sub_value.casefold() in self.empty_values:
                    continue
                if self.str_to_ideal_type:
                    yield key, Literal(string_to_ideal_type(sub_value))
                else:
                    yield key, Literal(sub_value)
            elif isinstance(sub_value, int | float | bool):
                yield key, Literal(sub_value)
            elif isinstance(sub_value, dict):
                yield from self._unpack_json(sub_value, key)
            elif isinstance(sub_value, list):
                for no, item in enumerate(sub_value, 1):
                    if isinstance(item, dict):
                        yield from self._unpack_json(item, f"{key}_{no}")
                    else:
                        yield from self._get_predicate_objects_pair(key, item, self.unpack_json)
            else:
                yield key, Literal(str(sub_value))


class DMSPropertyExtractor(DictExtractor):
    def __init__(
        self,
        id_: URIRef,
        data: Mapping[str, Any],
        namespace: Namespace,
        as_uri_ref: Callable[[Instance | dm.DirectRelationReference], URIRef],
        empty_values: Set[str] = DEFAULT_EMPTY_VALUES,
        str_to_ideal_type: bool = False,
        unpack_json: bool = False,
    ) -> None:
        super().__init__(id_, data, namespace, None, empty_values, str_to_ideal_type, unpack_json)
        self.as_uri_ref = as_uri_ref

    def _get_predicate_objects_pair(
        self, key: str, value: Any, unpack_json: bool
    ) -> Iterable[tuple[str, Literal | URIRef]]:
        if isinstance(value, dict) and "space" in value and "externalId" in value:
            yield key, self.as_uri_ref(dm.DirectRelationReference.load(value))
        else:
            yield from super()._get_predicate_objects_pair(key, value, unpack_json)



from collections.abc import Iterable, Sequence

from cognite.client import data_modeling as dm
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr
from rdflib import Namespace, URIRef

from cognite.neat._client import NeatClient
from cognite.neat._constants import COGNITE_SPACES, DEFAULT_NAMESPACE
from cognite.neat._issues import IssueList, NeatIssue, catch_warnings
from cognite.neat._issues.warnings import CDFAuthWarning, ResourceNotFoundWarning, ResourceRetrievalWarning
from cognite.neat._rules.importers import DMSImporter
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.models.data_types import Json
from cognite.neat._rules.models.entities import UnknownEntity
from cognite.neat._rules.models.information import InformationProperty
from cognite.neat._rules.transformers import DMSToInformation, VerifyDMSRules
from cognite.neat._shared import Triple

from ._base import KnowledgeGraphExtractor
from ._dms import DMSExtractor


class DMSGraphExtractor(KnowledgeGraphExtractor):
    def __init__(
        self,
        data_model: dm.DataModel[dm.View],
        client: NeatClient,
        namespace: Namespace = DEFAULT_NAMESPACE,
        issues: Sequence[NeatIssue] | None = None,
        instance_space: str | SequenceNotStr[str] | None = None,
        skip_cognite_views: bool = True,
        unpack_json: bool = False,
        str_to_ideal_type: bool = False,
    ) -> None:
        self._client = client
        self._data_model = data_model
        self._namespace = namespace or DEFAULT_NAMESPACE
        self._issues = IssueList(issues)
        self._instance_space = instance_space
        self._skip_cognite_views = skip_cognite_views
        self._unpack_json = unpack_json
        self._str_to_ideal_type = str_to_ideal_type

        self._views: list[dm.View] | None = None
        self._information_rules: InformationRules | None = None
        self._dms_rules: DMSRules | None = None

    @classmethod
    def from_data_model_id(
        cls,
        data_model_id: dm.DataModelIdentifier,
        client: NeatClient,
        namespace: Namespace = DEFAULT_NAMESPACE,
        instance_space: str | SequenceNotStr[str] | None = None,
        skip_cognite_views: bool = True,
        unpack_json: bool = False,
        str_to_ideal_type: bool = False,
    ) -> "DMSGraphExtractor":
        issues: list[NeatIssue] = []
        try:
            data_model = client.data_modeling.data_models.retrieve(data_model_id, inline_views=True)
        except CogniteAPIError as e:
            issues.append(CDFAuthWarning("retrieving data model", str(e)))
            return cls(
                cls._create_empty_model(dm.DataModelId.load(data_model_id)),
                client,
                namespace,
                issues,
                instance_space,
                skip_cognite_views,
                unpack_json,
                str_to_ideal_type,
            )
        if not data_model:
            issues.append(ResourceRetrievalWarning(frozenset({data_model_id}), "data model"))
            return cls(
                cls._create_empty_model(dm.DataModelId.load(data_model_id)),
                client,
                namespace,
                issues,
                instance_space,
                skip_cognite_views,
                unpack_json,
                str_to_ideal_type,
            )
        return cls(
            data_model.latest_version(),
            client,
            namespace,
            issues,
            instance_space,
            skip_cognite_views,
            unpack_json,
            str_to_ideal_type,
        )

    @classmethod
    def _create_empty_model(cls, data_model_id: dm.DataModelId) -> dm.DataModel:
        return dm.DataModel(
            data_model_id.space,
            data_model_id.external_id,
            data_model_id.version or "MISSING",
            is_global=False,
            last_updated_time=0,
            created_time=0,
            description=None,
            name=None,
            views=[],
        )

    @property
    def _model_views(self) -> list[dm.View]:
        if self._views is None:
            self._views = self._get_views()
        return self._views

    @property
    def description(self) -> str:
        return "Extracts a data model with nodes and edges."

    @property
    def source_uri(self) -> URIRef:
        space, external_id, version = self._data_model.as_id().as_tuple()
        return DEFAULT_NAMESPACE[f"{self._client.config.project}/{space}/{external_id}/{version}"]

    def extract(self) -> Iterable[Triple]:
        """Extracts the knowledge graph from the data model."""
        views = self._model_views
        if self._skip_cognite_views:
            views = [view for view in views if view.space not in COGNITE_SPACES]

        yield from DMSExtractor.from_views(
            self._client,
            views,
            instance_space=self._instance_space,
            unpack_json=self._unpack_json,
            str_to_ideal_type=self._str_to_ideal_type,
        ).extract()

    def _get_views(self) -> list[dm.View]:
        view_by_id: dict[dm.ViewId, dm.View] = {}
        if view_ids := [view_id for view_id in self._data_model.views if isinstance(view_id, dm.ViewId)]:
            try:
                # MyPy does not understand the isinstance check above.
                retrieved = self._client.data_modeling.views.retrieve(ids=view_ids)  #  type: ignore[arg-type]
            except CogniteAPIError as e:
                self._issues.append(CDFAuthWarning("retrieving views", str(e)))
            else:
                view_by_id.update({view.as_id(): view for view in retrieved})
        views: list[dm.View] = []
        data_model_id = self._data_model.as_id()
        for dm_view in self._data_model.views:
            if isinstance(dm_view, dm.View):
                views.append(dm_view)
            elif isinstance(dm_view, dm.ViewId):
                if view := view_by_id.get(dm_view):
                    views.append(view)
                else:
                    self._issues.append(ResourceNotFoundWarning(dm_view, "view", data_model_id, "data model"))
        return views

    def get_information_rules(self) -> InformationRules:
        """Returns the information rules that the extractor uses."""
        if self._information_rules is None:
            self._information_rules, self._dms_rules = self._create_rules()
        return self._information_rules

    def get_dms_rules(self) -> DMSRules:
        """Returns the DMS rules that the extractor uses."""
        if self._dms_rules is None:
            self._information_rules, self._dms_rules = self._create_rules()
        return self._dms_rules

    def get_issues(self) -> IssueList:
        """Returns the issues that occurred during the extraction."""
        return self._issues

    def _create_rules(self) -> tuple[InformationRules, DMSRules]:
        # The DMS and Information rules must be created together to link them property.
        importer = DMSImporter.from_data_model(self._client, self._data_model)
        unverified_dms = importer.to_rules()
        if self._unpack_json and (dms_rules := unverified_dms.rules):
            # Drop the JSON properties from the DMS rules as these are no longer valid.
            json_name = Json().name  # To avoid instantiating Json multiple times.
            dms_rules.properties = [
                prop
                for prop in dms_rules.properties
                if not (
                    (
                        isinstance(prop.value_type, Json)
                        or (isinstance(prop.value_type, str) and prop.value_type == json_name)
                    )
                    # We are not unpacking list of JSONs.
                    and prop.is_list is not True
                )
            ]

        with catch_warnings() as issues:
            # Any errors occur will be raised and caught outside the extractor.
            verified_dms = VerifyDMSRules(client=self._client).transform(unverified_dms)
            information_rules = DMSToInformation(self._namespace).transform(verified_dms)

        # We need to sync the metadata between the two rules, such that the `.sync_with_info_rules` method works.
        information_rules.metadata.physical = verified_dms.metadata.identifier
        verified_dms.metadata.logical = information_rules.metadata.identifier
        verified_dms.sync_with_info_rules(information_rules)

        # Adding startNode and endNode to the information rules for views that are used for edges.
        classes_by_prefix = {cls_.class_.prefix: cls_ for cls_ in information_rules.classes}
        for view in self._model_views:
            if view.used_for == "edge" and view.external_id in classes_by_prefix:
                cls_ = classes_by_prefix[view.external_id]
                for property_ in ("startNode", "endNode"):
                    information_rules.properties.append(
                        InformationProperty(
                            class_=cls_.class_,
                            property_=property_,
                            value_type=UnknownEntity(),
                            min_count=0,
                            max_count=1,
                        )
                    )

        self._issues.extend(issues)
        return information_rules, verified_dms



import re
import uuid
import xml.etree.ElementTree as ET
from functools import cached_property
from pathlib import Path
from typing import ClassVar
from typing import Literal as LiteralType
from xml.etree.ElementTree import Element

from rdflib import RDF, XSD, Literal, Namespace, URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._graph.extractors._base import BaseExtractor
from cognite.neat._issues.errors import FileReadError, NeatValueError
from cognite.neat._shared import Triple
from cognite.neat._utils.rdf_ import remove_namespace_from_uri
from cognite.neat._utils.text import to_camel_case
from cognite.neat._utils.xml_ import get_children

IODD = Namespace("http://www.io-link.com/IODD/2010/10/")
XSI = Namespace("http://www.w3.org/2001/XMLSchema-instance/")

XSI_XML_PREFIX = "{http://www.w3.org/2001/XMLSchema-instance}"


class IODDExtractor(BaseExtractor):
    """
    IODD-XML extractor of RDF triples

    Each IODD sheet describes an IODD device. This extractor extracts rdf triples that describes the device, and the
    sensors connected to the device.
    This data is described under the elements "DeviceIdentity" and "ProcessDataCollection".
    In addition, triples extacted from "DeviceIdentity" and
    "ProcessDataCollection" may reference "Text" elements which are found under "ExternalTextCollection". Edges to
    these Text element nodes are also extracted.

    Args:
        root: XML root element of IODD XML file.
        namespace: Optional custom namespace to use for extracted triples that define data
                    model instances. Defaults to DEFAULT_NAMESPACE.
        device_id: Optional user specified unique id/tag for actual equipment instance. If not provided, a randomly
        generated UUID will be used. The device_id must be WEB compliant,
        meaning that the characters /&?=: % are not allowed
    """

    device_elements_with_text_nodes: ClassVar[list[str]] = ["VendorText", "VendorUrl", "DeviceName", "DeviceFamily"]
    std_variable_elements_to_extract: ClassVar[list[str]] = ["V_SerialNumber", "V_ApplicationSpecificTag"]
    text_elements_language: LiteralType["en", "de"] = "en"

    def __init__(
        self,
        root: Element,
        namespace: Namespace | None = None,
        device_id: str | None = None,
    ):
        self.root = root
        self.namespace = namespace or DEFAULT_NAMESPACE

        if device_id and device_id != re.sub(r"[^a-zA-Z0-9-_.]", "", device_id):
            raise NeatValueError("Specified device_id is not web compliant. Please exclude characters: /&?=: %")

        self.device_id = (
            self.namespace[device_id] if device_id else self.namespace[f"Device_{str(uuid.uuid4()).replace('-', '_')}"]
        )

    @cached_property
    def _text_id_2value_mapping(self) -> dict[str, str]:
        """
        !!! note used for "Prototype Solution" !!!
        A mapping for text_id references to Text elements under ExternalTextCollection.
        The mapping can be used to find the Text element with matching id, and returns
        the value associated with the Text element.
        """
        mapping = {}
        if et_root := get_children(
            self.root, "ExternalTextCollection", ignore_namespace=True, include_nested_children=True, no_children=1
        ):
            if language_element := get_children(et_root[0], "PrimaryLanguage", ignore_namespace=True, no_children=1):
                if (
                    language_element[0].attrib.get("{http://www.w3.org/XML/1998/namespace}lang")
                    == self.text_elements_language
                ):
                    if text_elements := get_children(
                        language_element[0], child_tag="Text", ignore_namespace=True, include_nested_children=True
                    ):
                        for element in text_elements:
                            if id := element.attrib.get("id"):
                                if value := element.attrib.get("value"):
                                    mapping[id] = value
        return mapping

    @classmethod
    def from_file(cls, filepath: Path, namespace: Namespace | None = None, device_id: str | None = None):
        if filepath.suffix != ".xml":
            raise FileReadError(filepath, "File is not XML.")
        return cls(ET.parse(filepath).getroot(), namespace, device_id)

    @classmethod
    def _from_root2triples(cls, root: Element, namespace: Namespace, device_id: URIRef) -> list[Triple]:
        """Loops through the relevant elements of the IODD XML sheet to create rdf triples that describes the IODD
        device by starting at the root element.
        """
        triples: list[Triple] = []

        # Extract DeviceIdentity triples
        if di_root := get_children(
            root, "DeviceIdentity", ignore_namespace=True, include_nested_children=True, no_children=1
        ):
            triples.extend(cls._iodd_device_identity2triples(di_root[0], namespace, device_id))

        # Extract VariableCollection triples -
        # this element holds the information about the sensors connected to the device that collects data such as
        # temperature, voltage, leakage etc.
        if vc_root := get_children(
            root, "VariableCollection", ignore_namespace=True, include_nested_children=True, no_children=1
        ):
            triples.extend(cls._variables_data_collection2triples(vc_root[0], namespace, device_id))

        if pc_root := get_children(
            root, "ProcessDataCollection", ignore_namespace=True, include_nested_children=True, no_children=1
        ):
            triples.extend(cls._process_data_collection2triples(pc_root[0], namespace, device_id))

        if et_root := get_children(
            root, "ExternalTextCollection", ignore_namespace=True, include_nested_children=True, no_children=1
        ):
            triples.extend(cls._text_elements2triples(et_root[0], namespace))

        return triples

    @classmethod
    def _process_data_collection2triples(
        cls, pc_root: Element, namespace: Namespace, device_id: URIRef
    ) -> list[Triple]:
        """
        Will only collect ProcessDataIn elements at this point. The data from the IO-master is transmitted as an
        array related to a ProcessDataIn item.
        """
        triples: list[Triple] = []

        if process_data_in := get_children(
            pc_root, "ProcessDataIn", ignore_namespace=True, include_nested_children=True
        ):
            for process_data_element in process_data_in:
                if p_id := process_data_element.attrib.get("id"):
                    device_id_str = remove_namespace_from_uri(device_id)
                    process_data_in_id = namespace[f"{device_id_str}.{p_id}"]

                    # Create ProcessDataIn node
                    triples.append((process_data_in_id, RDF.type, IODD.ProcessDataIn))

                    # Create connection from device to node
                    triples.append((device_id, IODD.processDataIn, process_data_in_id))

                    # Connect record items (essentially an array of indexed variables) to the ProcessDataIn node
                    triples.extend(cls._process_data_in_records2triples(process_data_element, process_data_in_id))

        return triples

    @classmethod
    def _device_2text_elements_edges(cls, di_root: Element, id: URIRef, namespace: Namespace) -> list[Triple]:
        """
        Create edges from the device node to text nodes.
        """
        triples: list[Triple] = []

        for element_tag in cls.device_elements_with_text_nodes:
            if child := get_children(
                di_root, child_tag=element_tag, ignore_namespace=True, include_nested_children=True, no_children=1
            ):
                if text_id := child[0].attrib.get("textId"):
                    # Create connection from device to textId node
                    element_tag = to_camel_case(element_tag)
                    triples.append((id, IODD[element_tag], namespace[text_id]))

        return triples

    @classmethod
    def _text_elements2triples(cls, et_root: Element, namespace: Namespace) -> list[Triple]:
        """
        This method extracts all text item triples under the ExternalTextCollection element. This will create a node
        for each text item, and add the text value as a property to the node.
        """
        triples: list[Triple] = []

        if language_element := get_children(et_root, "PrimaryLanguage", ignore_namespace=True, no_children=1):
            if (
                language_element[0].attrib.get("{http://www.w3.org/XML/1998/namespace}lang")
                == cls.text_elements_language
            ):
                if text_elements := get_children(
                    language_element[0], child_tag="Text", ignore_namespace=True, include_nested_children=True
                ):
                    for element in text_elements:
                        if id := element.attrib.get("id"):
                            text_id = namespace[id]

                            # Create Text node
                            triples.append((text_id, RDF.type, IODD.TextObject))

                            # Resolve text value related to the text item
                            if value := element.attrib.get("value"):
                                triples.append((text_id, IODD.value, Literal(value)))
        return triples

    @classmethod
    def _std_variables2triples(cls, vc_root: Element, namespace: Namespace, device_id: URIRef) -> list[Triple]:
        """
        For simplicity, only extract the two items we want for this use case - V_ApplicationSpecificTag and
        V_SerialNumber
        """
        triples: list[Triple] = []

        if std_variable_elements := get_children(vc_root, child_tag="StdVariableRef", ignore_namespace=True):
            for element in std_variable_elements:
                if id := element.attrib.get("id"):
                    if id in cls.std_variable_elements_to_extract:
                        if object := element.attrib.get("defaultValue"):
                            predicate = to_camel_case(id.replace("V_", ""))
                            triples.append((device_id, IODD[predicate], Literal(object)))
        return triples

    @classmethod
    def _variables_data_collection2triples(
        cls, vc_root: Element, namespace: Namespace, device_id: URIRef
    ) -> list[Triple]:
        """
        VariableCollection contains elements that references Variables and StdVariables. The StdVariables
        can be resolved by looking up the ID in the IODD-StandardDefinitions1.1.xml sheet.

        The Variable elements are descriptions of the sensors collecting data for the device.
        """
        triples: list[Triple] = []

        # StdVariableRef elements of interest
        triples.extend(cls._std_variables2triples(vc_root, namespace, device_id))

        # Variable elements (these are the descriptions of the sensors)
        if variable_elements := get_children(vc_root, child_tag="Variable", ignore_namespace=True):
            for element in variable_elements:
                if id := element.attrib.get("id"):
                    device_id_str = remove_namespace_from_uri(device_id)
                    variable_id = f"{device_id_str}.{id}"

                    # Create connection from device node to time series
                    triples.append((device_id, IODD.variable, Literal(variable_id, datatype=XSD["timeseries"])))

        return triples

    @classmethod
    def _iodd_device_identity2triples(cls, di_root: Element, namespace: Namespace, device_id: URIRef) -> list[Triple]:
        """
        Properties and metadata related to the IO Device are described under the 'DeviceIdentity' element in the XML.
        This method extracts the triples that describe the device's identity which is found under the
        DeviceIdentity element and its child elements.

        """
        triples: list[Triple] = []

        # Create rdf type triple for IODD
        triples.append(
            (
                device_id,
                RDF.type,
                IODD.IoddDevice,
            )
        )

        for attribute_name, attribute_value in di_root.attrib.items():
            triples.append((device_id, IODD[attribute_name], Literal(attribute_value)))

        triples.extend(cls._device_2text_elements_edges(di_root, device_id, namespace))
        return triples

    @classmethod
    def _process_data_in_records2triples(cls, pc_in_root: Element, process_data_in_id: URIRef):
        """
        Extract RecordItems related to a ProcessDataIn element. Each record item is indexed. Will use this index
        as the identifier for the time series in CDF.
        """
        triples: list[Triple] = []

        if record_items := get_children(pc_in_root, "RecordItem", ignore_namespace=True, include_nested_children=True):
            for record in record_items:
                if index := record.attrib.get("subindex"):
                    process_id_str = remove_namespace_from_uri(process_data_in_id)
                    record_id = f"{process_id_str}.{index}"
                    # Create connection from device node to time series
                    triples.append((process_data_in_id, IODD.variable, Literal(record_id, datatype=XSD["timeseries"])))

        return triples

    def extract(self) -> list[Triple]:
        """
        Extract RDF triples from IODD XML
        """
        return self._from_root2triples(self.root, self.namespace, self.device_id)

    def _variable2info(self, variable_element: Element) -> dict:
        """
        !!! note used for "Prototype Solution" !!!
        Extracts information relevant to a CDF time series type from a Variable element
        """

        variable_dict = {}

        if name := get_children(
            variable_element, child_tag="Name", ignore_namespace=True, include_nested_children=False, no_children=1
        ):
            if text_id := name[0].get("textId"):
                variable_dict["name"] = self._text_id_2value_mapping[text_id]
        if description := get_children(
            variable_element,
            child_tag="Description",
            ignore_namespace=True,
            include_nested_children=False,
            no_children=1,
        ):
            if text_id := description[0].get("textId"):
                variable_dict["description"] = self._text_id_2value_mapping[text_id]
        if data_type := get_children(
            variable_element, child_tag="Datatype", ignore_namespace=True, include_nested_children=False, no_children=1
        ):
            variable_dict["data_type"] = data_type[0].attrib[f"{XSI_XML_PREFIX}type"]

        return variable_dict

    def _process_record2info(self, record_element: Element) -> dict:
        """
        !!! note used for "Prototype Solution" !!!
        Extracts information relevant to a CDF time series type from a Record element
        """
        record_dict = {}

        if name := get_children(
            record_element, child_tag="Name", ignore_namespace=True, include_nested_children=False, no_children=1
        ):
            if text_id := name[0].get("textId"):
                record_dict["name"] = self._text_id_2value_mapping[text_id]
        if description := get_children(
            record_element, child_tag="Description", ignore_namespace=True, include_nested_children=False, no_children=1
        ):
            if text_id := description[0].get("textId"):
                record_dict["description"] = self._text_id_2value_mapping[text_id]
        if data_type := get_children(
            record_element,
            child_tag="SimpleDatatype",
            ignore_namespace=True,
            include_nested_children=False,
            no_children=1,
        ):
            record_dict["data_type"] = data_type[0].attrib[f"{XSI_XML_PREFIX}type"]
        if index := record_element.attrib.get("subindex"):
            record_dict["index"] = index

        return record_dict

    def _extract_enhanced_ts_information(self, json_file_path: Path):
        """
        Extract additional information like name, description and data type for Variables and ProcessDataIn
        record elements in the IODD. The purpose is for the result gile to be used for enhancing time series with more
        information when they are created in CDF.

        Args:
            json_file_path: file path for where to write the extracted information about all time series
                            in the IODD

        !!! note "Prototype Solution" !!!
        This is an intermediate solution while better support for adding this information directly
        into the knowledge graph for the timeseries node type is under development.
        """
        import json

        ts_ext_id2_info_map = {}

        # Variable elements (these are the descriptions of the sensors)
        if variable_elements := get_children(
            self.root, child_tag="Variable", ignore_namespace=True, include_nested_children=True
        ):
            for element in variable_elements:
                if id := element.attrib.get("id"):
                    device_id_str = remove_namespace_from_uri(self.device_id)
                    variable_id = f"{device_id_str}.{id}"
                    ts_ext_id2_info_map[variable_id] = self._variable2info(element)

        if process_data_in := get_children(
            self.root, "ProcessDataIn", ignore_namespace=True, include_nested_children=True
        ):
            for process_data_element in process_data_in:
                if p_id := process_data_element.attrib.get("id"):
                    device_id_str = remove_namespace_from_uri(self.device_id)
                    process_data_in_id = f"{device_id_str}.{p_id}"
                    if record_items := get_children(
                        process_data_element, "RecordItem", ignore_namespace=True, include_nested_children=True
                    ):
                        for record in record_items:
                            if index := record.attrib.get("subindex"):
                                process_record_id = f"{process_data_in_id}.{index}"
                                ts_ext_id2_info_map[process_record_id] = self._process_record2info(record)

        with Path.open(json_file_path, "w") as fp:
            json.dump(ts_ext_id2_info_map, fp, indent=2)



from collections.abc import Iterable
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadata, FileMetadataFilter, FileMetadataList

from ._base import ClassicCDFBaseExtractor, InstanceIdPrefix


class FilesExtractor(ClassicCDFBaseExtractor[FileMetadata]):
    """Extract data from Cognite Data Fusions files metadata into Neat."""

    _default_rdf_type = "File"
    _instance_id_prefix = InstanceIdPrefix.file

    @classmethod
    def _from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
    ) -> tuple[int | None, Iterable[FileMetadata]]:
        items = client.files(data_set_external_ids=data_set_external_id)
        return None, items

    @classmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str
    ) -> tuple[int | None, Iterable[FileMetadata]]:
        total = client.files.aggregate(
            filter=FileMetadataFilter(asset_subtree_ids=[{"externalId": root_asset_external_id}])
        )[0].count
        items = client.files(asset_subtree_external_ids=root_asset_external_id)
        return total, items

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[FileMetadata]]:
        file_metadata = FileMetadataList.load(Path(file_path).read_text())
        return len(file_metadata), file_metadata



from abc import ABC
from dataclasses import dataclass
from typing import Any

from cognite.neat._shared import NeatList, NeatObject


@dataclass
class GraphTransformationResult(NeatObject, ABC):
    name: str
    affected_nodes_count: int | None = None
    added: int | None = None
    removed: int | None = None
    skipped: int | None = None
    modified: int | None = None

    def dump(self, aggregate: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {"name": self.name}
        if self.added:
            output["added"] = self.added
        if self.removed:
            output["removed"] = self.removed
        if self.skipped:
            output["skipped"] = self.skipped
        if self.affected_nodes_count:
            output["affected nodes"] = self.affected_nodes_count
        if self.modified:
            output["modified instances"] = self.modified
        return output


class GraphTransformationResultList(NeatList[GraphTransformationResult]):
    def _repr_html_(self) -> str:
        df = self.to_pandas().fillna(0)
        df = df.style.format({column: "{:,.0f}".format for column in df.select_dtypes(include="number").columns})
        return df._repr_html_()  # type: ignore[attr-defined]



from ._imf2rules import IMFImporter
from ._inference2rules import InferenceImporter, SubclassInferenceImporter
from ._owl2rules import OWLImporter

__all__ = ["IMFImporter", "InferenceImporter", "OWLImporter", "SubclassInferenceImporter"]



from collections.abc import Iterable
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import TimeSeries, TimeSeriesFilter, TimeSeriesList

from ._base import ClassicCDFBaseExtractor, InstanceIdPrefix


class TimeSeriesExtractor(ClassicCDFBaseExtractor[TimeSeries]):
    """Extract data from Cognite Data Fusions TimeSeries into Neat."""

    _default_rdf_type = "TimeSeries"
    _instance_id_prefix = InstanceIdPrefix.time_series

    @classmethod
    def _from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
    ) -> tuple[int | None, Iterable[TimeSeries]]:
        total = client.time_series.aggregate_count(
            filter=TimeSeriesFilter(data_set_ids=[{"externalId": data_set_external_id}])
        )
        items = client.time_series(data_set_external_ids=data_set_external_id)
        return total, items

    @classmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str
    ) -> tuple[int | None, Iterable[TimeSeries]]:
        total = client.time_series.aggregate_count(
            filter=TimeSeriesFilter(asset_subtree_ids=[{"externalId": root_asset_external_id}])
        )
        items = client.time_series(asset_subtree_external_ids=root_asset_external_id)
        return total, items

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[TimeSeries]]:
        timeseries = TimeSeriesList.load(Path(file_path).read_text())
        return len(timeseries), timeseries



from cognite.neat._client.data_classes.schema import DMSSchema

from ._rules import DMSContainer, DMSEnum, DMSMetadata, DMSNode, DMSProperty, DMSRules, DMSView
from ._rules_input import (
    DMSInputContainer,
    DMSInputEnum,
    DMSInputMetadata,
    DMSInputNode,
    DMSInputProperty,
    DMSInputRules,
    DMSInputView,
)
from ._validation import DMSValidation

__all__ = [
    "DMSContainer",
    "DMSEnum",
    "DMSInputContainer",
    "DMSInputEnum",
    "DMSInputMetadata",
    "DMSInputNode",
    "DMSInputProperty",
    "DMSInputRules",
    "DMSInputView",
    "DMSMetadata",
    "DMSNode",
    "DMSProperty",
    "DMSRules",
    "DMSSchema",
    "DMSValidation",
    "DMSView",
]



import json
import zipfile
from collections.abc import Iterable, Sequence
from pathlib import Path

from pydantic import ValidationError

from cognite.neat._issues import IssueList, MultiValueError, NeatIssue
from cognite.neat._issues.warnings import (
    FileItemNotSupportedWarning,
    FileMissingRequiredFieldWarning,
    FileReadWarning,
    FileTypeUnexpectedWarning,
    NeatValueWarning,
)
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.importers._base import BaseImporter
from cognite.neat._rules.importers._dtdl2rules.dtdl_converter import _DTDLConverter
from cognite.neat._rules.importers._dtdl2rules.spec import DTDL_CLS_BY_TYPE_BY_SPEC, DTDLBase, Interface
from cognite.neat._rules.models import InformationInputRules
from cognite.neat._rules.models.information import InformationInputMetadata
from cognite.neat._utils.text import humanize_collection, to_pascal_case


class DTDLImporter(BaseImporter[InformationInputRules]):
    """Importer from Azure Digital Twin - DTDL (Digital Twin Definition Language).

    This importer supports DTDL v2.0 and v3.0.

    It is recommended to use the class methods `from_directory` and `from_zip` to create an instance of this class.

    Args:
        items (Sequence[DTDLBase]): A sequence of DTDLBase objects.
        name (str, optional): Name of the data model. Defaults to None.
        read_issues (list[ValidationIssue], optional): A list of issues that occurred during reading. Defaults to None.
        schema (SchemaCompleteness, optional): Schema completeness. Defaults to SchemaCompleteness.partial.

    """

    def __init__(
        self,
        items: Sequence[DTDLBase],
        name: str | None = None,
        read_issues: list[NeatIssue] | None = None,
    ) -> None:
        self._items = items
        self.name = name
        self._read_issues = IssueList(read_issues)

    @classmethod
    def _from_file_content(cls, file_content: str, filepath: Path) -> Iterable[DTDLBase | NeatIssue]:
        raw = json.loads(file_content)
        if isinstance(raw, dict):
            if (context := raw.get("@context")) is None:
                yield FileMissingRequiredFieldWarning(filepath, "@context", "Missing '@context' key.")
                return
            raw_list = [raw]
        elif isinstance(raw, list):
            context = next(
                (entry["@context"] for entry in raw if isinstance(entry, dict) and "@context" in entry), None
            )
            if context is None:
                yield FileMissingRequiredFieldWarning(filepath, "@context", "Missing '@context' key.")
                return
            raw_list = raw
        else:
            yield FileTypeUnexpectedWarning(filepath, frozenset(["dict", "list"]), "Content is not an object or array.")
            return

        if isinstance(context, list):
            context = context[0]
        Interface.default_context = context
        spec_version = context.split(";")[1]
        try:
            cls_by_type = DTDL_CLS_BY_TYPE_BY_SPEC[spec_version]
        except KeyError:
            yield NeatValueWarning(
                f"Unsupported DTDL spec version: {spec_version} in {filepath}. "
                f"Supported versions are {humanize_collection(DTDL_CLS_BY_TYPE_BY_SPEC.keys())}."
                " The file will be skipped."
            )
            return

        for item in raw_list:
            if not (type_ := item.get("@type")):
                yield FileMissingRequiredFieldWarning(filepath, "@type", "Missing '@type' key.")
                continue
            cls_ = cls_by_type.get(type_)
            if cls_ is None:
                yield FileItemNotSupportedWarning(f"Unknown '@type' {type_}.", filepath=filepath)
                continue
            try:
                yield cls_.model_validate(item)
            except ValidationError as e:
                yield FileTypeUnexpectedWarning(filepath, frozenset([cls.__name__]), str(e))
            except Exception as e:
                yield FileReadWarning(filepath=filepath, reason=str(e))

    @classmethod
    def from_directory(cls, directory: Path) -> "DTDLImporter":
        items: list[DTDLBase] = []
        issues: list[NeatIssue] = []
        for filepath in directory.glob("**/*.json"):
            for item in cls._from_file_content(filepath.read_text(), filepath):
                if isinstance(item, NeatIssue):
                    issues.append(item)
                else:
                    items.append(item)
        return cls(items, directory.stem, read_issues=issues)

    @classmethod
    def from_zip(cls, zip_file: Path) -> "DTDLImporter":
        items: list[DTDLBase] = []
        issues: list[NeatIssue] = []
        with zipfile.ZipFile(zip_file) as z:
            for filepath in z.namelist():
                if filepath.endswith(".json"):
                    for item in cls._from_file_content(z.read(filepath).decode(), Path(filepath)):
                        if isinstance(item, NeatIssue):
                            issues.append(item)
                        else:
                            items.append(item)
        return cls(items, zip_file.stem, read_issues=issues)

    def to_rules(self) -> ReadRules[InformationInputRules]:
        converter = _DTDLConverter(self._read_issues)

        converter.convert(self._items)

        metadata = self._default_metadata()

        if self.name:
            metadata["name"] = to_pascal_case(self.name)
        try:
            most_common_prefix = converter.get_most_common_prefix()
        except ValueError:
            # No prefixes are defined so we just use the default prefix...
            ...
        else:
            metadata["space"] = most_common_prefix

        rules = InformationInputRules(
            metadata=InformationInputMetadata.load(metadata),
            properties=converter.properties,
            classes=converter.classes,
        )
        converter.issues.trigger_warnings()
        if converter.issues.has_errors:
            raise MultiValueError(converter.issues.errors)

        return ReadRules(rules, {})



import json
import warnings
from pathlib import Path
from typing import Literal, get_args

import yaml

from cognite.neat._rules._shared import VerifiedRules

from ._base import BaseExporter


class YAMLExporter(BaseExporter[VerifiedRules, str]):
    """Export rules (Information, DMS or Domain) to YAML.

    Args:
        files: The number of files to output. Defaults to "single".
        output: The format to output the rules. Defaults to "yaml".

    The following formats are available:

    - "single": A single YAML file will contain the entire rules.

    .. note::

        YAML files are typically used for storing rules when checked into version control systems, e.g., git-history.
        The advantage of using YAML files over Excel is that tools like git can show the differences between different
        versions of the rules.

    """

    Files = Literal["single"]
    Format = Literal["yaml", "json"]

    file_option = get_args(Files)
    format_option = get_args(Format)

    def __init__(self, files: Files = "single", output: Format = "yaml"):
        if files not in self.file_option:
            raise ValueError(f"Invalid files: {files}. Valid options are {self.file_option}")
        if output not in self.format_option:
            raise ValueError(f"Invalid output: {output}. Valid options are {self.format_option}")
        self.files = files
        self.output = output

    @property
    def description(self) -> str:
        return "Export verified model to YAML."

    def export_to_file(self, rules: VerifiedRules, filepath: Path) -> None:
        """Exports transformation rules to YAML/JSON file(s)."""
        if self.files == "single":
            if filepath.suffix != f".{self.output}":
                warnings.warn(f"File extension is not .{self.output}, adding it to the file name", stacklevel=2)
                filepath = filepath.with_suffix(f".{self.output}")
            filepath.write_text(self.export(rules), encoding=self._encoding, newline=self._new_line)
        else:
            raise NotImplementedError(f"Exporting to {self.files} files is not supported")

    def export(self, rules: VerifiedRules) -> str:
        """Export rules to YAML (or JSON) format.

        This will export the rules to YAML format if the output is set to "yaml" and JSON format if the output is set.
        All None and Unset values are excluded from the output to keep the output clean, i.e., only the values the user
        has set.

        Args:
            rules: The rules to be exported.

        Returns:
            str: The rules in YAML (or JSON) format.
        """
        # model_dump_json ensures that the output is in JSON format,
        # if we don't do this, we will get Enums and other types that are not serializable to YAML
        json_output = rules.dump(mode="json", sort=True, exclude_none=True, exclude_unset=True)
        if self.output == "json":
            return json.dumps(json_output)
        elif self.output == "yaml":
            return yaml.safe_dump(json_output)
        else:
            raise ValueError(f"Invalid output: {self.output}. Valid options are {self.format_option}")



from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any

from cognite.client.testing import CogniteClientMock

from cognite.neat._client._api_client import NeatClient

from ._api.data_modeling_loaders import DataModelLoaderAPI
from ._api.schema import SchemaAPI


class NeatClientMock(CogniteClientMock):
    """Mock for ToolkitClient object

    All APIs are replaced with specked MagicMock objects.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "parent" in kwargs:
            super().__init__(*args, **kwargs)
            return
        super().__init__(*args, **kwargs)
        self.schema = SchemaAPI(self)
        self.loaders = DataModelLoaderAPI(self)


@contextmanager
def monkeypatch_neat_client() -> Iterator[NeatClient]:
    neat_client_mock = NeatClientMock()
    NeatClient.__new__ = lambda *args, **kwargs: neat_client_mock  # type: ignore[method-assign]
    yield neat_client_mock
    NeatClient.__new__ = lambda cls, *args, **kwargs: object.__new__(cls)  # type: ignore[method-assign]



from __future__ import annotations

import itertools
import json
from datetime import datetime, timezone
from pathlib import Path
from types import GenericAlias
from typing import Any, ClassVar, Literal, cast, get_args

from openpyxl import Workbook
from openpyxl.cell import MergedCell
from openpyxl.styles import Alignment, Border, Font, PatternFill, Side
from openpyxl.worksheet.worksheet import Worksheet
from rdflib import Namespace

from cognite.neat._rules._constants import get_internal_properties
from cognite.neat._rules._shared import VerifiedRules
from cognite.neat._rules.models import (
    ExtensionCategory,
    SchemaCompleteness,
    SheetRow,
)
from cognite.neat._rules.models.dms import DMSMetadata
from cognite.neat._rules.models.information import InformationMetadata
from cognite.neat._rules.models.information._rules import InformationRules
from cognite.neat._utils.spreadsheet import find_column_with_value

from ._base import BaseExporter

MAX_COLUMN_WIDTH = 70.0


class ExcelExporter(BaseExporter[VerifiedRules, Workbook]):
    """Export rules to Excel.

    Args:
        styling: The styling to use for the Excel file. Defaults to "default". See below for details
            on the different styles.
        new_model_id: The new model ID to use for the exported spreadsheet. This is only applicable if the input
            rules have 'is_reference' set. If provided, the model ID will be used to automatically create the
            new metadata sheet in the Excel file. The model id is expected to be a tuple of (prefix, title)
            (space, external_id) for InformationRules and DMSRules respectively.

        sheet_prefix: The prefix to use for the sheet names in the Excel file. Defaults to an empty string.

    The following styles are available:

    - "none":    No styling is applied.
    - "minimal": Column widths are adjusted to fit the content, and the header row(s) is frozen.
    - "default": Minimal + headers are bold, increased size, and colored.
    - "maximal": Default + alternating row colors in the properties sheet for each class in addition to extra
                 blank rows between classes and borders
    """

    Style = Literal["none", "minimal", "default", "maximal"]
    DumpOptions = Literal["user", "last", "reference"]
    _main_header_by_sheet_name: ClassVar[dict[str, str]] = {
        "Properties": "Definition of Properties",
        "Classes": "Definition of Classes",
        "Views": "Definition of Views",
        "Containers": "Definition of Containers",
        "Nodes": "Definition of Nodes",
        "Enum": "Definition of Enum Collections",
    }
    style_options = get_args(Style)
    dump_options = get_args(DumpOptions)

    def __init__(
        self,
        styling: Style = "default",
        new_model_id: tuple[str, str] | None = None,
        sheet_prefix: str | None = None,
        reference_rules_with_prefix: tuple[VerifiedRules, str] | None = None,
        add_empty_rows: bool = False,
        hide_internal_columns: bool = True,
        include_properties: Literal["same-space", "all"] = "all",
    ):
        self.sheet_prefix = sheet_prefix or ""
        if styling not in self.style_options:
            raise ValueError(f"Invalid styling: {styling}. Valid options are {self.style_options}")
        self.styling = styling
        self._styling_level = self.style_options.index(styling)
        self.new_model_id = new_model_id
        self.reference_rules_with_prefix = reference_rules_with_prefix
        self.add_empty_rows = add_empty_rows
        self.hide_internal_columns = hide_internal_columns
        self.include_properties = include_properties

    @property
    def description(self) -> str:
        return "Export verified model to Excel."

    def export_to_file(self, rules: VerifiedRules, filepath: Path) -> None:
        """Exports transformation rules to excel file."""
        data = self.export(rules)
        try:
            data.save(filepath)
        finally:
            data.close()
        return None

    def export(self, rules: VerifiedRules) -> Workbook:
        workbook = Workbook()
        # Remove default sheet named "Sheet"
        workbook.remove(workbook["Sheet"])

        dumped_user_rules: dict[str, Any] = rules.dump(by_alias=True)

        self._write_metadata_sheet(workbook, dumped_user_rules["Metadata"], sheet_prefix=self.sheet_prefix)
        self._write_sheets(workbook, dumped_user_rules, rules, sheet_prefix=self.sheet_prefix)
        if self.reference_rules_with_prefix:
            reference_rules, prefix = self.reference_rules_with_prefix
            dumped_reference_rules = reference_rules.dump(entities_exclude_defaults=False, by_alias=True)
            self._write_sheets(workbook, dumped_reference_rules, reference_rules, sheet_prefix=prefix)
            self._write_metadata_sheet(workbook, dumped_reference_rules["Metadata"], sheet_prefix=prefix)

        if isinstance(rules, InformationRules) and rules.prefixes:
            self._write_prefixes_sheet(workbook, rules.prefixes)

        if self._styling_level > 0:
            self._adjust_column_widths(workbook)

        if self.hide_internal_columns:
            for sheet in workbook.sheetnames:
                if sheet.lower() == "metadata":
                    continue
                ws = workbook[sheet]
                for col in get_internal_properties():
                    column_letter = find_column_with_value(ws, col)
                    if column_letter:
                        ws.column_dimensions[column_letter].hidden = True

        return workbook

    def _write_sheets(
        self,
        workbook: Workbook,
        dumped_rules: dict[str, Any],
        rules: VerifiedRules,
        sheet_prefix: str = "",
    ):
        for sheet_name, headers in rules.headers_by_sheet(by_alias=True).items():
            if sheet_name in ("Metadata", "Prefixes", "Reference", "Last"):
                continue
            sheet = workbook.create_sheet(f"{sheet_prefix}{sheet_name}")

            main_header = self._main_header_by_sheet_name[sheet_name]
            sheet.append([main_header] + [""] * (len(headers) - 1))

            if headers[0] == "Neat ID":
                # Move the Neat ID to the end of the columns
                headers = headers[1:] + ["Neat ID"]

            sheet.append(headers)

            fill_colors = itertools.cycle(["CADCFC", "FFFFFF"])
            fill_color = next(fill_colors)
            last_class: str | None = None
            item: dict[str, Any]
            for item in dumped_rules.get(sheet_name) or []:
                if "Neat ID" in item:
                    # Move the Neat ID to the end of the columns
                    item["Neat ID"] = item.pop("Neat ID")
                row = list(item.values())
                class_ = row[0]

                is_properties = sheet_name == "Properties"
                is_new_class = class_ != last_class and last_class is not None
                if self._styling_level > 2 and is_new_class and is_properties:
                    if self.add_empty_rows:
                        sheet.append([""] * len(headers))
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                    fill_color = next(fill_colors)

                if is_properties and self.include_properties == "same-space":
                    space = class_.split(":")[0] if ":" in class_ else rules.metadata.space
                    if space != rules.metadata.space:
                        continue

                sheet.append(row)
                if self._styling_level > 2 and is_properties:
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                last_class = class_

            if self._styling_level > 0:
                # This freezes all rows above the given row
                sheet.freeze_panes = sheet["A3"]

                sheet["A1"].alignment = Alignment(horizontal="left")

            if self._styling_level > 1:
                # Make the header row bold, larger, and colored
                for cell, *_ in sheet.iter_cols(min_row=1, max_row=1, min_col=1, max_col=len(headers)):
                    cell.font = Font(bold=True, size=20)
                    cell.fill = PatternFill(fgColor="FFC000", patternType="solid")
                for cell in sheet["2"]:
                    cell.font = Font(bold=True, size=14)

    def _write_metadata_sheet(self, workbook: Workbook, metadata: dict[str, Any], sheet_prefix: str = "") -> None:
        # Excel does not support timezone in datetime strings
        if isinstance(metadata.get("created"), datetime):
            metadata["created"] = metadata["created"].replace(tzinfo=None)
        if isinstance(metadata.get("updated"), datetime):
            metadata["updated"] = metadata["updated"].replace(tzinfo=None)

        metadata_sheet = workbook.create_sheet(f"{sheet_prefix}Metadata")
        for key, value in metadata.items():
            metadata_sheet.append([key, value])

        if self._styling_level > 1:
            for cell in metadata_sheet["A"]:
                cell.font = Font(bold=True, size=12)

    def _write_prefixes_sheet(self, workbook: Workbook, prefixes: dict[str, Namespace]) -> None:
        metadata_sheet = workbook.create_sheet("Prefixes")
        metadata_sheet.append(["Prefix", "Namespace"])
        for key, value in prefixes.items():
            metadata_sheet.append([key, value])

        if self._styling_level > 1:
            for cell in metadata_sheet["A"]:
                cell.font = Font(bold=True, size=12)

    @classmethod
    def _get_item_class(cls, annotation: GenericAlias) -> type[SheetRow]:
        if not isinstance(annotation, GenericAlias):
            raise ValueError(f"Expected annotation to be a GenericAlias, but got {type(annotation)}")
        args = get_args(annotation)
        if len(args) != 1:
            raise ValueError(f"Expected annotation to have exactly one argument, but got {len(args)}")
        arg = args[0]
        if not issubclass(arg, SheetRow):
            raise ValueError(f"Expected annotation to have a BaseModel argument, but got {type(arg)}")
        return arg

    @classmethod
    def _adjust_column_widths(cls, workbook: Workbook) -> None:
        for sheet_ in workbook:
            sheet = cast(Worksheet, sheet_)
            for column_cells in sheet.columns:
                try:
                    max_length = max(len(str(cell.value)) for cell in column_cells if cell.value is not None)
                except ValueError:
                    max_length = 0

                selected_column = column_cells[0]
                if isinstance(selected_column, MergedCell):
                    selected_column = column_cells[1]

                current = sheet.column_dimensions[selected_column.column_letter].width or (max_length + 0.5)  # type: ignore[union-attr]
                sheet.column_dimensions[selected_column.column_letter].width = min(  # type: ignore[union-attr]
                    max(current, max_length + 0.5), MAX_COLUMN_WIDTH
                )
        return None


class _MetadataCreator:
    creator_name = "<YOUR NAME>"

    def __init__(
        self,
        action: Literal["create", "update"],
        new_model_id: tuple[str, str] | None = None,
    ):
        self.action = action
        self.new_model_id = new_model_id or ("YOUR_SPACE", "YOUR_EXTERNAL_ID")

    def create(self, metadata: InformationMetadata | DMSMetadata) -> dict[str, Any]:
        now = datetime.now(timezone.utc).replace(microsecond=0, tzinfo=None)
        if self.action == "update":
            output = json.loads(metadata.model_dump_json(by_alias=True))
            # This is the same for Information and DMS
            output["updated"] = now.isoformat()
            output["schema"] = SchemaCompleteness.extended.value
            output["extension"] = ExtensionCategory.addition.value
            if value := output.get("creator"):
                output["creator"] = f"{value}, {self.creator_name}"
            else:
                output["creator"] = self.creator_name
            return output

        new_metadata = self._create_new_info(now)
        if isinstance(metadata, DMSMetadata):
            from cognite.neat._rules.transformers._converters import _InformationRulesConverter

            output_metadata: DMSMetadata | InformationMetadata = _InformationRulesConverter._convert_metadata_to_dms(
                new_metadata
            )
        elif isinstance(metadata, InformationMetadata):
            output_metadata = new_metadata
        else:
            raise ValueError(f"Bug in Neat: Unknown metadata type: {type(metadata)}")

        created = json.loads(output_metadata.model_dump_json(by_alias=True))
        created.pop("extension", None)
        return created

    def _create_new_info(self, now: datetime) -> InformationMetadata:
        return InformationMetadata(
            space=self.new_model_id[0],
            external_id=self.new_model_id[1],
            description=None,
            version="1",
            created=now,
            updated=now,
            creator=[self.creator_name],
            name=self.new_model_id[1],
        )
